/****************************************************************************
 *                                                                          *
 * File    : hyperlink.c                                                    *
 *                                                                          *
 * Purpose : Hyperlink custom control functions.                            *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <commctrl.h>
#include <mmsystem.h>
#include "hyperlink.h"

#pragma comment(lib, "winmm.lib")

/** Extra message crackers **************************************************/

#define HANDLE_WM_CAPTURECHANGED(hwnd, wParam, lParam, fn) \
    ((fn)((hwnd), (HWND)(lParam)), 0)

/** Private message crackers ************************************************/

#define HANDLE_HLM_SETTOOLTIP(hwnd, wParam, lParam, fn) \
    ((fn)((hwnd), (LPTSTR)(LPARAM)(lParam)), 0)

#define HANDLE_HLM_GETTOOLTIP(hwnd, wParam, lParam, fn) \
    (LRESULT)(DWORD)(int)(fn)((hwnd), (int)(wParam), (LPTSTR)(lParam))

#define HANDLE_HLM_GETTOOLTIPLENGTH(hwnd, wParam, lParam, fn) \
    (LRESULT)(DWORD)(int)(fn)(hwnd)

#define HANDLE_HLM_SETFGCOLORS(hwnd, wParam, lParam, fn) \
    ((fn)((hwnd), (COLORREF)(WPARAM)(wParam), (COLORREF)(LPARAM)(lParam)), 0)

#define HANDLE_HLM_SETBGCOLORS(hwnd, wParam, lParam, fn) \
    ((fn)((hwnd), (COLORREF)(WPARAM)(wParam), (COLORREF)(LPARAM)(lParam)), 0)

/** Global variables ********************************************************/

typedef struct _HYPERLINK
{
    LPTSTR   szText;
    LPTSTR   szTooltip;
    HWND     hwndTooltip;
    COLORREF fgcNormal;
    COLORREF fgcHover;
    COLORREF bgcNormal;
    COLORREF bgcHover;
    BOOL     fHover;
}
HYPERLINK, *PHYPERLINK;

static PHYPERLINK pHyperlink = NULL;
static HFONT hFontWindow = NULL;
static LPTSTR szSoundPath = NULL;
static DWORD dwHypCtrlCount = 0;

/** Prototypes **************************************************************/

static LRESULT CALLBACK Hyperlink_WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
static BOOL Hyperlink_OnNcCreate(HWND, CREATESTRUCT *);
static void Hyperlink_OnNcDestroy(HWND);
static UINT Hyperlink_OnGetDlgCode(HWND, MSG *);
static void Hyperlink_OnMouseMove(HWND, int, int, UINT);
static void Hyperlink_OnCaptureChanged(HWND, HWND);
static void Hyperlink_OnWinIniChange(HWND, LPCTSTR);
static void Hyperlink_OnSize(HWND, UINT, int, int);
static void Hyperlink_OnLButtonDown(HWND, BOOL, int, int, UINT);
static void Hyperlink_OnLButtonUp(HWND, int, int, UINT);
static void Hyperlink_OnEnable(HWND, BOOL);
static BOOL Hyperlink_OnEraseBkgnd(HWND, HDC);
static void Hyperlink_OnSetFont(HWND, HFONT, BOOL);
static HFONT Hyperlink_OnGetFont(HWND);
static void Hyperlink_OnPaint(HWND);
static void Hyperlink_OnSetText(HWND, LPCTSTR);
static int Hyperlink_OnGetText(HWND, int, LPTSTR);
static int Hyperlink_OnGetTextLength(HWND);
static void Hyperlink_OnSetTooltip(HWND, LPCTSTR);
static int Hyperlink_OnGetTooltip(HWND, int, LPTSTR);
static int Hyperlink_OnGetTooltipLength(HWND);
static void Hyperlink_OnSetFgColors(HWND, COLORREF, COLORREF);
static void Hyperlink_OnSetBgColors(HWND, COLORREF, COLORREF);
static void Hyperlink_AdjustWindowSize(HWND);
static void Hyperlink_PreloadSound(void);

/** Macros ******************************************************************/

#define GWL_HYPERLINK 0

// this macro evaluates to the number of elements in an array
#define DIMOF(Array) (sizeof(Array) / sizeof(Array[0]))

// zero out a structure. if fInitSize is TRUE, initialize the first
// int to the size of the structure.
#define INITSTRUCT(structure, fInitSize) \
    (ZeroMemory(&(structure), sizeof(structure)), \
        fInitSize ? (*(int*) &(structure) = sizeof(structure)) : 0)

// simplify memory handling
#define HEAPALLOC(size) HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, size)
#define HEAPREALLOC(mem, size) HeapReAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, mem, size)
#define HEAPSIZE(mem) HeapSize(GetProcessHeap(), 0, mem)
#define HEAPFREE(mem) if (mem != NULL) { HeapFree(GetProcessHeap(), 0, mem); mem = NULL; }

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_RegisterClass                                        *
 *                                                                          *
 * Purpose : Register the hyperlink control window class.                   *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

BOOL Hyperlink_RegisterClass(void)
{
    WNDCLASS wc;

    INITSTRUCT(wc, FALSE);
    wc.style         = CS_GLOBALCLASS;
    wc.lpfnWndProc   = (WNDPROC) Hyperlink_WndProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = sizeof(PHYPERLINK);
    wc.hInstance     = GetModuleHandle(NULL);
    wc.hbrBackground = NULL;
    wc.lpszMenuName  = NULL;
    wc.hIcon         = NULL;
    wc.hCursor       = LoadCursor(NULL, IDC_HAND);
    wc.lpszClassName = HYPERLINK_CLASS;

    if (!RegisterClass(&wc)) return FALSE;

    return TRUE;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_UnregisterClass                                      *
 *                                                                          *
 * Purpose : Unregister the hyperlink control window class.                 *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

BOOL Hyperlink_UnregisterClass(void)
{
    return UnregisterClass(HYPERLINK_CLASS, GetModuleHandle(NULL));
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_WndProc                                              *
 *                                                                          *
 * Purpose : Hyperlink control window procedure.                            *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static LRESULT CALLBACK Hyperlink_WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg)
    {
        // windows messages
        HANDLE_MSG(hwnd, WM_NCCREATE, Hyperlink_OnNcCreate);
        HANDLE_MSG(hwnd, WM_NCDESTROY, Hyperlink_OnNcDestroy);
        HANDLE_MSG(hwnd, WM_GETDLGCODE, Hyperlink_OnGetDlgCode);
        HANDLE_MSG(hwnd, WM_MOUSEMOVE, Hyperlink_OnMouseMove);
        HANDLE_MSG(hwnd, WM_CAPTURECHANGED, Hyperlink_OnCaptureChanged);
        HANDLE_MSG(hwnd, WM_WININICHANGE, Hyperlink_OnWinIniChange);
        HANDLE_MSG(hwnd, WM_SIZE, Hyperlink_OnSize);
        HANDLE_MSG(hwnd, WM_LBUTTONDOWN, Hyperlink_OnLButtonDown);
        HANDLE_MSG(hwnd, WM_LBUTTONUP, Hyperlink_OnLButtonUp);
        HANDLE_MSG(hwnd, WM_ENABLE, Hyperlink_OnEnable);
        HANDLE_MSG(hwnd, WM_ERASEBKGND, Hyperlink_OnEraseBkgnd);
        HANDLE_MSG(hwnd, WM_PAINT, Hyperlink_OnPaint);
        HANDLE_MSG(hwnd, WM_SETFONT, Hyperlink_OnSetFont);
        HANDLE_MSG(hwnd, WM_GETFONT, Hyperlink_OnGetFont);
        HANDLE_MSG(hwnd, WM_SETTEXT, Hyperlink_OnSetText);
        HANDLE_MSG(hwnd, WM_GETTEXT, Hyperlink_OnGetText);
        HANDLE_MSG(hwnd, WM_GETTEXTLENGTH, Hyperlink_OnGetTextLength);

        // private messages
        HANDLE_MSG(hwnd, HLM_SETTOOLTIP, Hyperlink_OnSetTooltip);
        HANDLE_MSG(hwnd, HLM_GETTOOLTIP, Hyperlink_OnGetTooltip);
        HANDLE_MSG(hwnd, HLM_GETTOOLTIPLENGTH, Hyperlink_OnGetTooltipLength);
        HANDLE_MSG(hwnd, HLM_SETFGCOLORS, Hyperlink_OnSetFgColors);
        HANDLE_MSG(hwnd, HLM_SETBGCOLORS, Hyperlink_OnSetBgColors);

        default:
            return DefWindowProc(hwnd, msg, wParam, lParam);
    }
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnNcCreate                                           *
 *                                                                          *
 * Purpose : Handle WM_NCCREATE message.                                    *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static BOOL Hyperlink_OnNcCreate(HWND hwnd, CREATESTRUCT *pcs)
{
    if ((pHyperlink = HEAPALLOC(sizeof(HYPERLINK))) != NULL)
    {
        hFontWindow = GetWindowFont(pcs->hwndParent);

        ZeroMemory(pHyperlink, sizeof(HYPERLINK));

        pHyperlink->szText      = NULL;
        pHyperlink->szTooltip   = NULL;
        pHyperlink->hwndTooltip = NULL;
        pHyperlink->fgcNormal   = RGB(0, 0, 128);
        pHyperlink->fgcHover    = RGB(128, 0, 0);
        pHyperlink->bgcNormal   = (COLORREF) GetSysColor(COLOR_3DFACE);
        pHyperlink->bgcHover    = (COLORREF) GetSysColor(COLOR_3DFACE);
        pHyperlink->fHover      = FALSE;

        SetWindowLong(hwnd, GWL_HYPERLINK, (LONG) pHyperlink);

        if (szSoundPath == NULL) Hyperlink_PreloadSound();

        Hyperlink_SetText(hwnd, pcs->lpszName);

        dwHypCtrlCount++;

        return TRUE;
    }

    return FALSE;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnNcDestroy                                          *
 *                                                                          *
 * Purpose : Handle WM_NCDESTROY message.                                   *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnNcDestroy(HWND hwnd)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    Hyperlink_SetTooltip(hwnd, NULL);

    HEAPFREE(pHyperlink->szText);
    HEAPFREE(pHyperlink);

    pHyperlink = NULL;

    dwHypCtrlCount--;
    if (dwHypCtrlCount == 0) HEAPFREE(szSoundPath);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnGetDlgCode                                         *
 *                                                                          *
 * Purpose : Handle WM_GETDLGCODE message.                                  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static UINT Hyperlink_OnGetDlgCode(HWND hwnd, MSG *pmsg)
{
    return DLGC_STATIC;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnMouseMove                                          *
 *                                                                          *
 * Purpose : Handle WM_MOUSEMOVE message.                                   *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnMouseMove(HWND hwnd, int x, int y, UINT keyFlags)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    if (pHyperlink->fHover)
    {
        POINT ptMousePos;
        RECT rc;

        ptMousePos.x = x;
        ptMousePos.y = y;

        ScreenToClient(hwnd, &ptMousePos);
        GetClientRect(hwnd, &rc);

        if (!PtInRect(&rc, ptMousePos))
        {
            pHyperlink->fHover = FALSE;
            ReleaseCapture();
        }
    }
    else
    {
        if (GetCapture() != hwnd)
        {
            pHyperlink->fHover = TRUE;
            SetCapture(hwnd);
        }
    }
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnCaptureChanged                                     *
 *                                                                          *
 * Purpose : Handle WM_CAPTURECHANGED message.                              *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnCaptureChanged(HWND hwnd, HWND hwndCtrl)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);
    pHyperlink->fHover = FALSE;
    InvalidateRect(hwnd, NULL, FALSE);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnWinIniChange                                       *
 *                                                                          *
 * Purpose : Handle WM_WININICHANGE message.                                *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnWinIniChange(HWND hwnd, LPCTSTR lpszSectionName)
{
    InvalidateRect(hwnd, NULL, FALSE);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnSize                                               *
 *                                                                          *
 * Purpose : Handle WM_SIZE message.                                        *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnSize(HWND hwnd, UINT state, int cx, int cy)
{
    DWORD dwStyle;

    dwStyle = GetWindowLong(hwnd, GWL_STYLE);
    if (dwStyle & HLS_AUTOSIZE) Hyperlink_AdjustWindowSize(hwnd);
    InvalidateRect(hwnd, NULL, FALSE);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnLButtonDown                                        *
 *                                                                          *
 * Purpose : Handle WM_LBUTTONDOWN message.                                 *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnLButtonDown(HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags)
{
    DWORD dwStyle;

    SetFocus(hwnd);

    dwStyle = GetWindowLong(hwnd, GWL_STYLE);
    if (dwStyle & HLS_PLAYCLICK)
    {
        PlaySound(szSoundPath, NULL, SND_ASYNC | SND_NODEFAULT | SND_NOWAIT);
    }
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnLButtonUp                                          *
 *                                                                          *
 * Purpose : Handle WM_LBUTTONUP message.                                   *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnLButtonUp(HWND hwnd, int x, int y, UINT keyFlags)
{
    FORWARD_WM_COMMAND(GetParent(hwnd), GetWindowID(hwnd), hwnd, HLN_CLICKED, SNDMSG);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnEnable                                             *
 *                                                                          *
 * Purpose : Handle WM_ENABLE message.                                      *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnEnable(HWND hwnd, BOOL fEnable)
{
    if (IsWindowEnabled(hwnd) != fEnable)
    {
        pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);
        pHyperlink->fHover = FALSE;
        InvalidateRect(hwnd, NULL, FALSE);
    }
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnEraseBkgnd                                         *
 *                                                                          *
 * Purpose : Handle WM_ERASEBKGND message.                                  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static BOOL Hyperlink_OnEraseBkgnd(HWND hwnd, HDC hdc)
{
    return TRUE;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnSetFont                                            *
 *                                                                          *
 * Purpose : Handle WM_SETFONT message.                                     *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnSetFont(HWND hwnd, HFONT hfont, BOOL fRedraw)
{
    DWORD dwStyle;

    hFontWindow = hfont;

    dwStyle = GetWindowLong(hwnd, GWL_STYLE);
    if (dwStyle & HLS_AUTOSIZE) Hyperlink_AdjustWindowSize(hwnd);
    if (fRedraw) InvalidateRect(hwnd, NULL, FALSE);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnGetFont                                            *
 *                                                                          *
 * Purpose : Handle WM_GETFONT message.                                     *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static HFONT Hyperlink_OnGetFont(HWND hwnd)
{
    return hFontWindow;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnPaint                                              *
 *                                                                          *
 * Purpose : Handle WM_PAINT message.                                       *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnPaint(HWND hwnd)
{
    DWORD dwStyle;
    UINT dtFlags = DT_SINGLELINE | DT_NOPREFIX;
    HFONT hFontText;
    LOGFONT lf;
    PAINTSTRUCT ps;
    RECT rc;
    HDC hdcMem;
    HBITMAP hbmMem, hbmOld;

    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    dwStyle = GetWindowLong(hwnd, GWL_STYLE);

    if (dwStyle & HLS_LEFT) dtFlags |= DT_LEFT;
    else if (dwStyle & HLS_HCENTER) dtFlags |= DT_CENTER;
    else if (dwStyle & HLS_RIGHT) dtFlags |= DT_RIGHT;

    if (dwStyle & HLS_TOP) dtFlags |= DT_TOP;
    else if (dwStyle & HLS_VCENTER) dtFlags |= DT_VCENTER;
    else if (dwStyle & HLS_BOTTOM) dtFlags |= DT_BOTTOM;

    ZeroMemory(&lf, sizeof(LOGFONT));
    GetObject(hFontWindow, sizeof(LOGFONT), &lf);

    BeginPaint(hwnd, &ps);

    SetMapMode(ps.hdc, MM_TEXT);
    SetTextAlign(ps.hdc, TA_LEFT | TA_TOP | TA_NOUPDATECP);

    GetClientRect(hwnd, &rc);

    hdcMem = CreateCompatibleDC(ps.hdc);
    hbmMem = CreateCompatibleBitmap(ps.hdc, rc.right, rc.bottom);
    hbmOld = SelectBitmap(hdcMem, hbmMem);

    if (pHyperlink->fHover)
    {
        FillRect(hdcMem, &rc, CreateSolidBrush(pHyperlink->bgcHover));

        if ((dwStyle & HLS_ULHOVER) || (dwStyle & HLS_ULALWAYS))
        {
            lf.lfUnderline = (BYTE) TRUE;
            hFontText = CreateFontIndirect(&lf);
        }
        else
        {
            hFontText = CreateFontIndirect(&lf);
        }

        SetBkColor(hdcMem, pHyperlink->bgcHover);
        SetTextColor(hdcMem, pHyperlink->fgcHover);
    }
    else
    {
        FillRect(hdcMem, &rc, CreateSolidBrush(pHyperlink->bgcNormal));

        if ((dwStyle & HLS_ULNONE) || (dwStyle & HLS_ULHOVER))
        {
            hFontText = CreateFontIndirect(&lf);
        }
        else
        {
            lf.lfUnderline = (BYTE) TRUE;
            hFontText = CreateFontIndirect(&lf);
        }

        SetBkColor(hdcMem, pHyperlink->bgcNormal);
        SetTextColor(hdcMem, pHyperlink->fgcNormal);
    }

    if (dwStyle & WS_DISABLED) SetTextColor(hdcMem, GetSysColor(COLOR_GRAYTEXT));

    SelectFont(hdcMem, hFontText);
    DrawText(hdcMem, pHyperlink->szText, -1, &rc, dtFlags);

    BitBlt(ps.hdc, 0, 0, rc.right, rc.bottom, hdcMem, 0, 0, SRCCOPY);

    DeleteFont(hFontText);
    SelectBitmap(hdcMem, hbmOld);
    DeleteBitmap(hbmMem);
    DeleteDC(hdcMem);

    EndPaint(hwnd, &ps);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnSetText                                            *
 *                                                                          *
 * Purpose : Handle WM_SETTEXT message.                                     *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnSetText(HWND hwnd, LPCTSTR lpszText)
{
    DWORD dwStyle;

    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    HEAPFREE(pHyperlink->szText);

    pHyperlink->szText = HEAPALLOC(lstrlen(lpszText) * sizeof(TCHAR) + sizeof(TCHAR));
    if (pHyperlink->szText == NULL) return;
    CopyMemory(pHyperlink->szText, lpszText, lstrlen(lpszText) * sizeof(TCHAR) + sizeof(TCHAR));

    dwStyle = GetWindowLong(hwnd, GWL_STYLE);
    if (dwStyle & HLS_AUTOSIZE) Hyperlink_AdjustWindowSize(hwnd);

    InvalidateRect(hwnd, NULL, FALSE);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnGetText                                            *
 *                                                                          *
 * Purpose : Handle WM_GETTEXT message.                                     *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int Hyperlink_OnGetText(HWND hwnd, int cchTextMax, LPTSTR lpszText)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    lstrcpyn(lpszText, pHyperlink->szText, cchTextMax);

    return lstrlen(lpszText);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnGetTextLength                                      *
 *                                                                          *
 * Purpose : Handle WM_GETTEXTLENGTH message.                               *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int Hyperlink_OnGetTextLength(HWND hwnd)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    return lstrlen(pHyperlink->szText);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnSetTooltip                                         *
 *                                                                          *
 * Purpose : Handle HLM_SETTOOLTIP message.                                 *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnSetTooltip(HWND hwnd, LPCTSTR lpszText)
{
    TOOLINFO ti;

    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    INITSTRUCT(ti, TRUE);
    ti.uFlags = TTF_SUBCLASS;
    ti.hwnd   = hwnd;
    ti.uId    = GetWindowID(hwnd);
    GetClientRect(hwnd, &ti.rect);

    if (lpszText == NULL)
    {
        ToolTip_DelTool(pHyperlink->hwndTooltip, &ti);
        pHyperlink->hwndTooltip = NULL;
        HEAPFREE(pHyperlink->szTooltip);
        return;
    }

    if (pHyperlink->hwndTooltip != NULL)
    {
        pHyperlink->szTooltip = HEAPREALLOC(pHyperlink->szTooltip, lstrlen(lpszText) * sizeof(TCHAR) + sizeof(TCHAR));
        if (pHyperlink->szTooltip == NULL) return;

        CopyMemory(pHyperlink->szTooltip, lpszText, lstrlen(lpszText) * sizeof(TCHAR) + sizeof(TCHAR));

        ti.lpszText = pHyperlink->szTooltip;
        ToolTip_UpdateTipText(pHyperlink->hwndTooltip, &ti);
    }
    else
    {
        pHyperlink->hwndTooltip = CreateWindowEx(WS_EX_TOPMOST | WS_EX_TOOLWINDOW,
            TOOLTIPS_CLASS,
            NULL,
            WS_POPUP,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            CW_USEDEFAULT,
            hwnd,
            NULL,
            (HINSTANCE) GetWindowLong(hwnd, GWL_HINSTANCE),
            NULL);

        if (pHyperlink->hwndTooltip == NULL) return;

        SetWindowPos(pHyperlink->hwndTooltip,
            HWND_TOPMOST,
            0, 0, 0, 0,
            SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

        pHyperlink->szTooltip = HEAPALLOC(lstrlen(lpszText) * sizeof(TCHAR) + sizeof(TCHAR));
        if (pHyperlink->szTooltip == NULL) return;

        CopyMemory(pHyperlink->szTooltip, lpszText, lstrlen(lpszText) * sizeof(TCHAR) + sizeof(TCHAR));

        ti.lpszText = pHyperlink->szTooltip;
        ToolTip_AddTool(pHyperlink->hwndTooltip, &ti);
    }
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnGetTooltip                                         *
 *                                                                          *
 * Purpose : Handle HLM_GETTOOLTIP message.                                 *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int Hyperlink_OnGetTooltip(HWND hwnd, int cchTextMax, LPTSTR lpszText)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    lstrcpyn(lpszText, pHyperlink->szTooltip, cchTextMax);

    return lstrlen(lpszText);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnGetTooltipLength                                   *
 *                                                                          *
 * Purpose : Handle HLM_GETTOOLTIPLENGTH message.                           *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int Hyperlink_OnGetTooltipLength(HWND hwnd)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    return lstrlen(pHyperlink->szTooltip);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnSetFgColors                                        *
 *                                                                          *
 * Purpose : Handle HLM_SETFGCOLORS message.                                *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnSetFgColors(HWND hwnd, COLORREF fgcNormal, COLORREF fgcHover)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    if ((pHyperlink->fgcNormal != fgcNormal) || (pHyperlink->fgcHover != fgcHover))
    {
        pHyperlink->fgcNormal = fgcNormal;
        pHyperlink->fgcHover = fgcHover;
        InvalidateRect(hwnd, NULL, FALSE);
    }
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnSetBgColors                                        *
 *                                                                          *
 * Purpose : Handle HLM_SETBGCOLORS message.                                *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_OnSetBgColors(HWND hwnd, COLORREF bgcNormal, COLORREF bgcHover)
{
    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    if ((pHyperlink->bgcNormal != bgcNormal) || (pHyperlink->bgcHover != bgcHover))
    {
        pHyperlink->bgcNormal = bgcNormal;
        pHyperlink->bgcHover = bgcHover;
        InvalidateRect(hwnd, NULL, FALSE);
    }
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_AdjustWindowSize                                     *
 *                                                                          *
 * Purpose : Adjust window size to fit the text.                            *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_AdjustWindowSize(HWND hwnd)
{
    PAINTSTRUCT ps;
    HDC hdc;
    SIZE size;

    pHyperlink = (PHYPERLINK) GetWindowLong(hwnd, GWL_HYPERLINK);

    hdc = BeginPaint(hwnd, &ps);

    SelectFont(hdc, hFontWindow);
    GetTextExtentPoint32(hdc, pHyperlink->szText, lstrlen(pHyperlink->szText), &size);

    EndPaint(hwnd, &ps);

    SetWindowPos(hwnd, 0, 0, 0, size.cx, size.cy, SWP_NOMOVE | SWP_NOZORDER);
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_PreloadSound                                         *
 *                                                                          *
 * Purpose : Load system navigating sound.                                  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static void Hyperlink_PreloadSound(void)
{
    LPCTSTR szRegPath = _T("AppEvents\\Schemes\\Apps\\Explorer\\Navigating\\.Current");
    HKEY hKey = NULL;
    DWORD dwSize = 0;
    LPTSTR szBuffer = NULL;

    if (RegOpenKeyEx(HKEY_CURRENT_USER, szRegPath, 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS)
    {
        RegQueryValueEx(hKey, NULL, 0, NULL, (LPBYTE) szBuffer, &dwSize);
        szBuffer = HEAPALLOC(dwSize);

        if (szBuffer != NULL)
        {
            if (RegQueryValueEx(hKey, NULL, 0, NULL, (LPBYTE) szBuffer, &dwSize) == ERROR_SUCCESS)
            {
                dwSize = ExpandEnvironmentStrings(szBuffer, NULL, 0);
                szSoundPath = HEAPALLOC(dwSize * sizeof(TCHAR) + sizeof(TCHAR));

                if (szSoundPath != NULL)
                {
                    ExpandEnvironmentStrings(szBuffer, szSoundPath, HEAPSIZE(szSoundPath) / sizeof(TCHAR));
                }
            }
        }
    }

    if (hKey != NULL) RegCloseKey(hKey);
}
