/****************************************************************************
 *                                                                          *
 * File    : main.c                                                         *
 *                                                                          *
 * Purpose : Pelles C IDE interface.                                        *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#include <commctrl.h>
#include "main.h"

/** Global variables ********************************************************/

CCSTYLEFLAGA aHyperlinkStyleFlags[] =
{
    { HLS_LEFT     , 0, "HLS_LEFT"      },
    { HLS_HCENTER  , 0, "HLS_HCENTER"   },
    { HLS_RIGHT    , 0, "HLS_RIGHT"     },
    { HLS_TOP      , 0, "HLS_TOP"       },
    { HLS_VCENTER  , 0, "HLS_VCENTER"   },
    { HLS_BOTTOM   , 0, "HLS_BOTTOM"    },
    { HLS_ULALWAYS , 0, "HLS_ULALWAYS"  },
    { HLS_ULNONE   , 0, "HLS_ULNONE"    },
    { HLS_ULHOVER  , 0, "HLS_ULHOVER"   },
    { HLS_PLAYCLICK, 0, "HLS_PLAYCLICK" },
    { HLS_AUTOSIZE , 0, "HLS_AUTOSIZE"  }
};

static HANDLE ghAsCstmCtrl = NULL;
static LPCCSTYLE gpccsHyperlink;

/** Prototypes **************************************************************/

static BOOL CALLBACK Hyperlink_CustomizeStyle(HWND, LPCCSTYLE);
static BOOL CALLBACK Hyperlink_DlgProc(HWND, UINT, WPARAM, LPARAM);
static BOOL Hyperlink_OnInitDialog(HWND, HWND, LPARAM);
static BOOL Hyperlink_OnCommand(HWND, int, HWND, UINT);
static int CALLBACK Hyperlink_SizeToText(DWORD, DWORD, HFONT, LPSTR);

/** Macros ******************************************************************/

// this macro evaluates to the number of elements in an array
#define DIMOF(Array) (sizeof(Array) / sizeof(Array[0]))

// zero out a structure. if fInitSize is TRUE, initialize the first
// int to the size of the structure.
#define INITSTRUCT(structure, fInitSize) \
    (ZeroMemory(&(structure), sizeof(structure)), \
        fInitSize ? (*(int*) &(structure) = sizeof(structure)) : 0)

// simplify memory handling
#define HEAPALLOC(size) HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, size)
#define HEAPREALLOC(mem, size) HeapReAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY, mem, size)
#define HEAPSIZE(mem) HeapSize(GetProcessHeap(), 0, mem)
#define HEAPFREE(mem) if (mem != NULL) { HeapFree(GetProcessHeap(), 0, mem); mem = NULL; }

/****************************************************************************
 *                                                                          *
 * Function: DllMain                                                        *
 *                                                                          *
 * Purpose : DLL entry and exit procedure.                                  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

BOOL WINAPI DllMain(HANDLE hDLL, DWORD dwReason, LPVOID lpReserved)
{
    ghAsCstmCtrl = hDLL;

    switch (dwReason)
    {
        case DLL_PROCESS_ATTACH:
            InitCommonControls();
            if (!Hyperlink_RegisterClass()) return FALSE;
            return TRUE;
        case DLL_PROCESS_DETACH:
            if (!Hyperlink_UnregisterClass()) return FALSE;
            return TRUE;
        default:
            return TRUE;
    }
}

/****************************************************************************
 *                                                                          *
 * Function: CustomControlInfoA                                             *
 *                                                                          *
 * Purpose : Custom control interface function. This is basically what      *
 *           makes the DLL a custom control.                                *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

UINT __declspec(dllexport) CALLBACK CustomControlInfoA(LPCCINFOA acci)
{
    if (!acci) return 2;

    INITSTRUCT(acci[0], FALSE);

    acci[0].cxDefault         = 80;
    acci[0].cyDefault         = 8;
    acci[0].flStyleDefault    = WS_CHILD | WS_VISIBLE | WS_TABSTOP | HLS_STANDARD;
    acci[0].cStyleFlags       = 11;
    acci[0].aStyleFlags       = aHyperlinkStyleFlags;
    acci[0].lpfnStyle         = Hyperlink_CustomizeStyle;
    acci[0].lpfnSizeToText    = Hyperlink_SizeToText;

    lstrcpyn(acci[0].szClass, HYPERLINK_CLASS, CCHCCCLASS);
    lstrcpyn(acci[0].szDesc, _T("Hyperlink (hyperlink.dll)"), CCHCCDESC);
    lstrcpyn(acci[0].szTextDefault, _T("Hyperlink"), CCHCCTEXT);

    return 1;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_CustomizeStyle                                       *
 *                                                                          *
 * Purpose : Invoke the custom control styles dialog.                       *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static BOOL CALLBACK Hyperlink_CustomizeStyle(HWND hwndParent, LPCCSTYLE pccs)
{
    int ret;

    gpccsHyperlink = pccs;

    ret = DialogBox(ghAsCstmCtrl, MAKEINTRESOURCE(DLG_HYPERLINK), hwndParent, Hyperlink_DlgProc);
    if (ret == -1)
    {
        MessageBox(hwndParent,
            _T("Hyperlink_CustomizeStyle(): DialogBox failed!"),
            _T("ascstmctrl.dll"),
            MB_OK | MB_ICONEXCLAMATION | MB_APPLMODAL);

        ret = 0;
    }

    return (BOOL) ret;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_DlgProc                                              *
 *                                                                          *
 * Purpose : Hyperlink control style dialog procedure.                      *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static BOOL CALLBACK Hyperlink_DlgProc(HWND hwndDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg)
    {
        HANDLE_DLGMSG(hwndDlg, WM_INITDIALOG, Hyperlink_OnInitDialog);
        HANDLE_DLGMSG(hwndDlg, WM_COMMAND, Hyperlink_OnCommand);
    }

    return FALSE;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnInitDialog                                         *
 *                                                                          *
 * Purpose : Handle WM_INITDIALOG message.                                  *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static BOOL Hyperlink_OnInitDialog(HWND hwndDlg, HWND hwndFocus, LPARAM lParam)
{
    HWND hwndTemp;

    hwndTemp = GetDlgItem(hwndDlg, ID_HALIGN);
    ComboBox_AddString(hwndTemp, _T("HLS_LEFT"));
    ComboBox_AddString(hwndTemp, _T("HLS_HCENTER"));
    ComboBox_AddString(hwndTemp, _T("HLS_RIGHT"));

    if (gpccsHyperlink->flStyle & HLS_LEFT)
        ComboBox_SetCurSel(hwndTemp, 0);
    else if (gpccsHyperlink->flStyle & HLS_HCENTER)
        ComboBox_SetCurSel(hwndTemp, 1);
    else if (gpccsHyperlink->flStyle & HLS_RIGHT)
        ComboBox_SetCurSel(hwndTemp, 2);

    hwndTemp = GetDlgItem(hwndDlg, ID_VALIGN);
    ComboBox_AddString(hwndTemp, _T("HLS_TOP"));
    ComboBox_AddString(hwndTemp, _T("HLS_VCENTER"));
    ComboBox_AddString(hwndTemp, _T("HLS_BOTTOM"));

    if (gpccsHyperlink->flStyle & HLS_TOP)
        ComboBox_SetCurSel(hwndTemp, 0);
    else if (gpccsHyperlink->flStyle & HLS_VCENTER)
        ComboBox_SetCurSel(hwndTemp, 1);
    else if (gpccsHyperlink->flStyle & HLS_BOTTOM)
        ComboBox_SetCurSel(hwndTemp, 2);

    hwndTemp = GetDlgItem(hwndDlg, ID_UNDERLINE);
    ComboBox_AddString(hwndTemp, _T("HLS_ULALWAYS"));
    ComboBox_AddString(hwndTemp, _T("HLS_ULNONE"));
    ComboBox_AddString(hwndTemp, _T("HLS_ULHOVER"));

    if (gpccsHyperlink->flStyle & HLS_ULALWAYS)
        ComboBox_SetCurSel(hwndTemp, 0);
    else if (gpccsHyperlink->flStyle & HLS_ULNONE)
        ComboBox_SetCurSel(hwndTemp, 1);
    else if (gpccsHyperlink->flStyle & HLS_ULHOVER)
        ComboBox_SetCurSel(hwndTemp, 2);

    if (gpccsHyperlink->flStyle & HLS_PLAYCLICK)
        CheckDlgButton(hwndDlg, ID_PLAYCLICK, BST_CHECKED);

    if (gpccsHyperlink->flStyle & HLS_AUTOSIZE)
        CheckDlgButton(hwndDlg, ID_AUTOSIZE, BST_CHECKED);

    return TRUE;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_OnCommand                                            *
 *                                                                          *
 * Purpose : Handle WM_COMMAND message.                                     *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static BOOL Hyperlink_OnCommand(HWND hwndDlg, int id, HWND hwndCtl, UINT codeNotify)
{
    HWND hwndTemp;

    switch (id)
    {
        case ID_DEFAULTS:

            hwndTemp = GetDlgItem(hwndDlg, ID_HALIGN);
            ComboBox_SetCurSel(hwndTemp, 0);

            hwndTemp = GetDlgItem(hwndDlg, ID_VALIGN);
            ComboBox_SetCurSel(hwndTemp, 0);

            hwndTemp = GetDlgItem(hwndDlg, ID_UNDERLINE);
            ComboBox_SetCurSel(hwndTemp, 2);

            CheckDlgButton(hwndDlg, ID_PLAYCLICK, BST_CHECKED);

            break;
        case IDCANCEL:

            EndDialog(hwndDlg, FALSE);

            break;
        case IDOK:

            hwndTemp = GetDlgItem(hwndDlg, ID_HALIGN);
            gpccsHyperlink->flStyle &= ~(HLS_LEFT | HLS_HCENTER | HLS_RIGHT);
            switch (ComboBox_GetCurSel(hwndTemp))
            {
                case 0:
                    gpccsHyperlink->flStyle |= HLS_LEFT;
                    break;
                case 1:
                    gpccsHyperlink->flStyle |= HLS_HCENTER;
                    break;
                case 2:
                    gpccsHyperlink->flStyle |= HLS_RIGHT;
                    break;
            }

            hwndTemp = GetDlgItem(hwndDlg, ID_VALIGN);
            gpccsHyperlink->flStyle &= ~(HLS_TOP | HLS_VCENTER | HLS_BOTTOM);
            switch (ComboBox_GetCurSel(hwndTemp))
            {
                case 0:
                    gpccsHyperlink->flStyle |= HLS_TOP;
                    break;
                case 1:
                    gpccsHyperlink->flStyle |= HLS_VCENTER;
                    break;
                case 2:
                    gpccsHyperlink->flStyle |= HLS_BOTTOM;
                    break;
            }

            hwndTemp = GetDlgItem(hwndDlg, ID_UNDERLINE);
            gpccsHyperlink->flStyle &= ~(HLS_ULALWAYS | HLS_ULNONE | HLS_ULHOVER);
            switch (ComboBox_GetCurSel(hwndTemp))
            {
                case 0:
                    gpccsHyperlink->flStyle |= HLS_ULALWAYS;
                    break;
                case 1:
                    gpccsHyperlink->flStyle |= HLS_ULNONE;
                    break;
                case 2:
                    gpccsHyperlink->flStyle |= HLS_ULHOVER;
                    break;
            }

            if (IsDlgButtonChecked(hwndDlg, ID_PLAYCLICK))
                gpccsHyperlink->flStyle |= HLS_PLAYCLICK;
            else
                gpccsHyperlink->flStyle &= ~HLS_PLAYCLICK;

            if (IsDlgButtonChecked(hwndDlg, ID_AUTOSIZE))
                gpccsHyperlink->flStyle |= HLS_AUTOSIZE;
            else
                gpccsHyperlink->flStyle &= ~HLS_AUTOSIZE;

            EndDialog(hwndDlg, TRUE);

            break;
    }

    return TRUE;
}

/****************************************************************************
 *                                                                          *
 * Function: Hyperlink_SizeToText                                           *
 *                                                                          *
 * Purpose : Custom control "Size to text" function.                        *
 *                                                                          *
 * History : Date      Reason                                               *
 *           00/00/00  Created                                              *
 *                                                                          *
 ****************************************************************************/

static int CALLBACK Hyperlink_SizeToText(DWORD flStyle, DWORD flExtStyle, HFONT hFont, LPSTR pszText)
{
    return -1;
}
